@@ -275,6 +275,9 @@ def get_build_environ(build_system):
         if env_var in os.environ:
             env[env_var] = os.environ[env_var]
 
+    # filter out 'None' values
+    env = {k: v for (k, v) in env.items() if v is not None}
+
     return env
 
 
@@ -378,9 +381,8 @@ def ssh_no_host_key_check():
 
 def setup_parser_common_args(parser):
     parser.add_argument('config',
-                        help='Config file, using .config.yaml if none is '
-                        'specified and using the current directory as '
-                        'repository anchor',
+                        help='Config file, using .config.yaml in KAS_WORK_DIR '
+                        'if none is specified',
                         nargs='?')
     parser.add_argument('--skip',
                         help='Skip build steps',
@@ -391,3 +393,30 @@ def setup_parser_common_args(parser):
     parser.add_argument('--update', action='store_true',
                         help='Pull new upstream changes to the desired '
                         'refspec even if it is already checked out locally')
+
+
+def setup_parser_preserve_env_arg(parser):
+    parser.add_argument('-E', '--preserve-env',
+                        help='Keep current user environment block',
+                        action='store_true')
+
+
+def run_handle_preserve_env_arg(ctx, os, args, SetupHome):
+    if args.preserve_env:
+        # Warn if there's any settings that setup_home would apply
+        # but are now ignored
+        for var in SetupHome.ENV_VARS:
+            if var in os.environ:
+                logging.warning('Environment variable "%s" ignored '
+                                'because user environment is being used',
+                                var)
+
+        if not os.isatty(sys.stdout.fileno()):
+            logging.error("Error: --preserve-env can only be "
+                          "run from a tty")
+            sys.exit(1)
+
+        ctx.environ = os.environ.copy()
+
+        logging.warning("Preserving the current environment block may "
+                        "have unintended side effects on the build.")
